;; Sieve of Eratosthenes using a large vp-set for marks  File:  oddsv.lisp
;; This version omits even integers.
;; Parameters set for simulator

(*proclaim '(type fixnum B V))

(defparameter B 32)  ;; Default vp-set size
(defparameter V  4)  ;; Mark bit vp-set has B*V vp's

;; Vp-set for mark bits

(def-vp-set bitvp (list B V))

(*proclaim '(type Boolean-pvar mrk))

(*defvar mrk t!! "marks" bitvp)  ;; Marks for n >= B+2

;; mrk = t ==> number is not factored by any prime so far

(*proclaim '(type Boolean-pvar bmk prm))

(*defvar bmk t!!)    ;; Marks for 2 <= n <= B+1, t ==> not yet factored
(*defvar prm nil!!)  ;; Integer is prime, prm ==> not bmk

(*proclaim '(type (signed-pvar 32) n mp jx mx))

(*defvar n              ;; Integer rep'd by P
         (+!! (ash!! (self-address!!) (!! 1)) (!! 3)))
(*defvar mp (!! 0))
(*defvar jx (!! 0))
(*defvar mx (!! 0))

(*proclaim '(type fixnum pnx pna stp 2B+1 2BV+2B+1 BV Br q phx prn pht))

(defvar pnx      0)  ;; Most recently identified prime
(defvar pna      0)  ;; Address of pnx
(defvar stp      0)  ;; B*pnx
(defvar 2B+1     0)  ;; Largest number in P = 2*B + 1 
(defvar 2BV+2B+1 0)  ;; Largest number represented = 2*B*V + 2*B + 1
(defvar BV       0)  ;; Address limit for marks = B*V
(defvar Br       1)  ;; V/P for default VP set
(defvar q        0)  ;; (BV+B+1) div pnx
(defvar phx      0)  ;; Number of sieve phases
(defvar prn    0)  ;; Number of primes sieved
(defvar pht    0)  ;; Total number of mrk phases

(defun setup ()
 (*cold-boot :initial-dimensions (list B))
 (setq pnx 0 prn 0 pht 0)
 (setq 2B+1 (1+ (ash B 1)))
 (setq 2BV+2B+1 (+ (ash (* B V) 1) 2B+1))
 (setq Br (truncate B 32))  ;; FOR SIMULATOR
 (setq BV (* B V))
 (list 'Br Br '2B+1 2B+1 '2BV+2B+1 2BV+2B+1 'BV BV)
)

;; Identify next prime <= B+1 and sieve

(defun sieve1 ()
  (*when bmk (setq pnx (*min n)))  ;; Identify next prime
  (*set mp (*!! n (!! pnx)))
  (*set jx (ash!! (-!! mp (!! 3)) (!! -1))) ;; Proc addr.
  (*set mx  (-!! jx (!! B)))                ;; Addr in bitvp
  (*when bmk
    (*if (<=!! mp (!! 2B+1))
        (*pset :no-collisions nil!! bmk jx) ;; in P's
        (*when (<!! mx (!! BV))             ;; in M's
          (*pset :no-collisions nil!! mrk mx :vp-set bitvp))))
  (setq stp (* B pnx)) (setq q (truncate 2BV+2B+1 pnx))
  (setq phx (truncate (- q 3) (ash B 1)))
  (dotimes (i phx)  ;; Loop to finish marking off M's
    (*set mx (+!! mx (!! stp)))  ;; Address of next mark
    (*when (<!! mx (!! BV))      ;; Mark off more numbers
      (*pset :no-collisions nil!! mrk mx :vp-set bitvp))
  ) ; end dotimes
  (setq pna (ash (- pnx 3) -1))  ;; Update processor for pnx
  (*setf (pref prm pna) t)  (*setf (pref bmk pna) nil)
  (incf prn) (incf pht phx)
)

;; Count total number of primes (assumes sieving is complete)

(*proclaim '(type fixnum prmcount))

(defvar prmcount 0)   ;; Count of total number of primes

;; Count primes, including 2 (implicit)

(defun count-primes ()
  (*if (or!! prm bmk) (*set jx (!! 1)) (*set jx (!! 0)))
  (*set mx (ash!! (self-address!!)
           (!! (the (signed-byte 6) (logcount (1- V))))))
  (dotimes (j V) ;; Tally primes in vp's with same phys. addr
    (*if (pref!! mrk mx :vp-set bitvp :collision-mode :no-collisions)
	 (*set jx (1+!! jx)))
    (*set mx (1+!! mx))) ;; end dotimes
  (setq prmcount (1+ (*sum jx)))
)

;; Finish sieve - may be called at any time, assumes
;; 2BV+2B+1 > 4.

(*proclaim '(type fixnum reportf))

(defvar reportf 2)  ;;Frequency of reports.

(defun finish-sieve ()
  (prog (pl (pc 0)) (declare (type fixnum pl pc))
   LOOP (setq pl (+ pnx 2))
        (when (> (* pl pl) 2BV+2B+1)
          (format t
           "Finished at ~D, ~D primes sieved, ~D mark-bit phases~%"
                        pnx prn               pht)
          (return 'FINISHED))
        (sieve1)
        (when (= (incf pc) reportf)
          (setq pc 0) (format t "~7D processed~%" pnx))
        (go loop)
  )
)

;; Report performance for last call to sieve1 parameter is
;; elapsed time.

(defun rx (etm)
  (declare (type single-float etm))
  (let* ((pn (* Br (1+ phx))) (pps (/ pn etm)) (mspp (/ (* 1000 etm) pn)))
    (format t
      "prime: ~6D, ~3D phases, ~6,2F ph/sec, ~8,3F msec/ph~%"
              pnx   pn          pps           mspp))
)
